/*
    
  This file is a part of Shellp, a 'shell helper' library.
  
  Copyright (C) 2011-2012 Jori Liesenborgs
  
  Contact: jori.liesenborgs@gmail.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  
  USA

*/

#include "shellconfig.h"
#include "shellinstance.h"
#include "shellcmdarg.h"
#include "shellcommand.h"
#include "iosystem.h"
#include "ncursesiosystem.h"
#include "fileiosystem.h"
#include "stdiosystem.h"
#include "lineinput.h"
#ifdef SHELLP_CONFIG_QT4
	#include "shellmainwin.h"
	#include "qtiosystem.h"
	#include <QApplication>
#endif // SHELLP_CONFIG_QT4
#ifdef SHELLP_CONFIG_NCURSES
	#include <signal.h>
#endif // SHELLP_CONFIG_NCURSES
#include "muparser/muParser.h"

#include <stdlib.h>
#include <iostream>
#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif // WIN32

#include "debugnew.h"

using namespace jthread;

namespace shellp
{

ShellInstance *ShellInstance::m_pInstance = 0;

ShellInstance::ShellInstance()
{
	if (m_pInstance != 0)
	{
		std::cerr << "Trying to create a second instance of ShellInstance, but only one is allowed!" << std::endl;
		exit(-1);
	}

	m_pInstance = this;

	if (m_mutex.Init() != 0)
	{
		std::cerr << "Fatal error: unable to initialize instance mutex" << std::endl;
		exit(-1);
	}

	m_pIOSys = 0;
	m_interactive = false;
	m_requestExit = false;
	m_errExit = false;

	m_pMuParser = new mu::Parser();
	m_pMuParser->ClearConst();
}

ShellInstance::~ShellInstance()
{
	clearVariables();
	delete m_pMuParser;

	// TODO: check this
	m_pInstance = 0;
}

ShellInstance *ShellInstance::getInstance()
{
	if (m_pInstance == 0)
	{
		std::cerr << "ShellInstance::getInstance() was called but no instance currently exists" << std::endl;
		exit(-1);
	}
	return m_pInstance;
}

void ShellInstance::registerCommand(ShellCommand *pCmd, const std::string &cmdName)
{
	// first, we'll see if the command isn't already registered
	
	for (int i = 0 ; i < m_commandList.size() ; i++)
	{
		if (m_commandList[i].getName() == cmdName)
		{
			std::string errStr = std::string("ERROR: command name'") + cmdName + std::string("' is already registered!");
			unrecoverableError(errStr);
		}
	}

	// ok, register new command
	
	m_commandList.push_back(CommandName(pCmd, cmdName));
}
 
void ShellInstance::unregisterCommand(ShellCommand *pCmd)
{
	// remove references to cmd

	std::vector<CommandName> newArray;

	for (int i = 0 ; i < m_commandList.size() ; i++)
	{
		if (m_commandList[i].getCommand() != pCmd)
			newArray.push_back(CommandName(m_commandList[i].getCommand(), m_commandList[i].getName()));
	}

	m_commandList = newArray;
}
	
void ShellInstance::unrecoverableError(const std::string &errStr)
{
	if (m_pIOSys)
		m_pIOSys->writeOutputLine(errStr);
	else
		std::cerr << errStr << std::endl;
#ifndef WIN32
	sleep(2);
#else
	Sleep(2000);
#endif // WIN32
	exit(-1);
}

bool ShellInstance::processCommandLine(const std::vector<std::string> &cmdLine)
{
	if (cmdLine.size() == 0) // no command specified
		return true;
	
	std::string cmdName = cmdLine[0];
	bool found = false;
	int i = 0;
	
	while(!found && i < m_commandList.size())
	{
		if (m_commandList[i].getName() == cmdName)
			found = true;
		else
			i++;
	}
	if (!found)
	{
		m_pIOSys->writeOutputLine(std::string("Unrecognized command: ") + cmdName);
		return false;
	}
	
	std::vector<std::string> args;

	for (int j = 1 ; j < cmdLine.size() ; j++)
		args.push_back(cmdLine[j]);

	ShellCommand *pCmd = m_commandList[i].getCommand();
	std::vector<ShellCmdArg *> cmdArgs = pCmd->getArguments();
	std::string errStr;
	
	if (!ShellCmdArg::parseArgumentList(cmdArgs, args, errStr))
	{
		m_pIOSys->writeOutputLine("Error parsing argument list:");
		m_pIOSys->writeOutputLine(errStr);
		m_pIOSys->writeOutputLine("");
		m_pIOSys->writeOutputLine("Usage:");
		m_pIOSys->writeOutputLine(cmdName + std::string(" ") + ShellCmdArg::getArgumentNames(cmdArgs));
		m_pIOSys->writeOutputLine("");
		return false;
	}

	// ok, arguments are processed, now execute the command

	if (!pCmd->execute())
	{
		m_pIOSys->writeOutputLine(std::string("Error executing command: ") + cmdName);
		m_pIOSys->writeOutputLine(pCmd->getErrorString());
		return false;
	}
	
	return true;
}

std::vector<std::string> ShellInstance::getCommandNames() const
{
	std::vector<std::string> names;
	
	for (int i = 0 ; i < m_commandList.size() ; i++)
		names.push_back(m_commandList[i].getName());

	return names;
}

ShellCommand *ShellInstance::getCommand(const std::string &commandName)
{
	for (int i = 0 ; i < m_commandList.size() ; i++)
	{
		if (m_commandList[i].getName() == commandName)
			return m_commandList[i].getCommand();
	}
	return 0;
}

void ShellInstance::appendCommandLine(const std::string &line)
{
	m_commandLines.push_back(line);
}

std::vector<std::string> ShellInstance::getCommandLines() const
{
	return m_commandLines;
}

/*
mu::value_type* ShellInstance::addParserVariable(const mu::char_type *a_szName, void *a_pUserData)
{
	ShellInstance *pInst = (ShellInstance *)a_pUserData;
	mu::value_type *pVal = new mu::value_type;

	*pVal = 0;
	pInst->m_autoVariables.push_back(pVal);

	return pVal;
}
*/

bool ShellInstance::evaluateExpression(const std::string &expression, double &value)
{
	mu::value_type result = 0;

	try
	{
		m_pMuParser->SetExpr(expression);
		result = m_pMuParser->Eval();
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot handle expression: ") + e.GetMsg());
		return false;
	}
	value = (double)result;
	return true;
}

bool ShellInstance::storeVariable(const std::string &name, double value)
{
	std::map<std::string, void *>::iterator it;

	it = m_variables.find(name);
	if (it != m_variables.end())
	{
		mu::value_type *pVar = (mu::value_type *)m_variables[name];
		*pVar = value;
	}
	else
	{
		mu::value_type *pVar = new mu::value_type;
		
		try
		{
			*pVar = value;
			m_pMuParser->DefineVar(name, pVar);
		}
		catch (mu::Parser::exception_type &e)
		{
			delete pVar;
			setErrorString(std::string("Math parser cannot handle variable name: ") + e.GetMsg());
			return false;
		}
		m_variables[name] = (void *)pVar;
	}

	return true;
}

void ShellInstance::getParserNamesAndValues(std::vector<std::string> &constNames, std::vector<double> &constValues,
			             std::vector<std::string> &varNames, std::vector<double> &varValues)
{
	constNames.resize(0);
	constValues.resize(0);
	varNames.resize(0);
	varValues.resize(0);

	mu::valmap_type cMap = m_pMuParser->GetConst();
	mu::varmap_type vMap = m_pMuParser->GetVar();
	mu::valmap_type::const_iterator item;
	mu::varmap_type::const_iterator item2;

	for (item = cMap.begin() ; item != cMap.end() ; item++)
	{
		constNames.push_back(item->first);
		constValues.push_back(item->second);
	}
	for (item2 = vMap.begin() ; item2 != vMap.end() ; item2++)
	{
		varNames.push_back(item2->first);
		varValues.push_back(*(item2->second));
	}
}

void ShellInstance::clearVariables()
{
	m_pMuParser->ClearVar();

	std::map<std::string, void *>::iterator it;

	for (it = m_variables.begin() ; it != m_variables.end() ; it++)
	{
		mu::value_type *pVar = (mu::value_type *)(it->second);
		delete pVar;
	}
	m_variables.clear();
}

bool ShellInstance::setInterrupted(bool f)
{
	m_mutex.Lock();
	m_interrupted = f;
	m_mutex.Unlock();
	return true;
}

bool ShellInstance::isInterrupted()
{
	bool value;

	m_mutex.Lock();
	value = m_interrupted;
	m_mutex.Unlock();

	return value;
}

bool ShellInstance::registerUnitName(int unitID, const std::string &name)
{
	if (unitID < 0)
	{
		setErrorString("Only positive unit IDs are allowed");
		return false;
	}

	std::map<int, std::pair<std::string, double> >::iterator it = m_units.find(unitID);
	if (it != m_units.end())
	{
		setErrorString("Unit ID has already been registered");
		return false;
	}

	m_units[unitID] = std::pair<std::string, double>(name, 1);

	return true;
}

bool ShellInstance::setUnitMultiplier(int unitID, double value)
{
	std::map<int, std::pair<std::string, double> >::iterator it = m_units.find(unitID);
	if (it == m_units.end())
	{
		setErrorString("Unit ID has not been registered");
		return false;
	}

	m_units[unitID].second = value;

	return true;
}

bool ShellInstance::getUnitName(int unitID, std::string &name) const
{
	std::map<int, std::pair<std::string, double> >::const_iterator it = m_units.find(unitID);
	if (it == m_units.end())
	{
		setErrorString("Unit ID has not been registered");
		return false;
	}

	name = it->second.first;

	return true;
}

double ShellInstance::getUnitMultiplier(int unitID) const
{
	std::map<int, std::pair<std::string, double> >::const_iterator it = m_units.find(unitID);
	if (it == m_units.end())
	{
		setErrorString("Unit ID has not been registered");
		return false;
	}

	return it->second.second;
}

bool ShellInstance::defineConstant(const std::string &name, double value)
{
	try
	{
		m_pMuParser->DefineConst(name.c_str(), value);
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot define specified constant: ") + e.GetMsg());
		return false;
	}
	return true;
}

void ShellInstance::clearConstants()
{
	m_pMuParser->ClearConst();
}

bool ShellInstance::defineFunction(const std::string &name, FunctionWith1Arg callBack)
{
	try
	{
		m_pMuParser->DefineFun(name.c_str(), callBack, false);
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot define specified function: ") + e.GetMsg());
		return false;
	}
	return true;
}

bool ShellInstance::defineFunction(const std::string &name, FunctionWith2Arg callBack)
{
	try
	{
		m_pMuParser->DefineFun(name.c_str(), callBack, false);
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot define specified function: ") + e.GetMsg());
		return false;
	}
	return true;
}

bool ShellInstance::defineFunction(const std::string &name, FunctionWith3Arg callBack)
{
	try
	{
		m_pMuParser->DefineFun(name.c_str(), callBack, false);
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot define specified function: ") + e.GetMsg());
		return false;
	}
	return true;
}

bool ShellInstance::defineFunction(const std::string &name, FunctionWith4Arg callBack)
{
	try
	{
		m_pMuParser->DefineFun(name.c_str(), callBack, false);
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot define specified function: ") + e.GetMsg());
		return false;
	}
	return true;
}

bool ShellInstance::defineFunction(const std::string &name, FunctionWith5Arg callBack)
{
	try
	{
		m_pMuParser->DefineFun(name.c_str(), callBack, false);
	}
	catch (mu::Parser::exception_type &e)
	{
		setErrorString(std::string("Math parser cannot define specified function: ") + e.GetMsg());
		return false;
	}
	return true;
}

bool ShellInstance::registerPreprocessorVariable(const std::string &name, const std::string &contents)
{
	std::string::size_type pos = name.find_first_not_of(ShellCmdArg::getValidPreprocessorNameCharacters());
	if (pos != std::string::npos)
	{
		setErrorString("The specified preprocessor variable name contains at least one illegal character");
		return false;
	}

	m_preprocessorVariables[name] = contents;
	return true;
}

bool ShellInstance::getPreprocessorVariableContent(const std::string &name, std::string &contents) const
{
	std::map<std::string, std::string>::const_iterator it;

	it = m_preprocessorVariables.find(name);
	if (it == m_preprocessorVariables.end())
	{
		setErrorString("Specified name is not a registered preprocessor variable name");
		return false;
	}
	
	contents = it->second;

	return true;
}

void ShellInstance::getPreprocessorVariables(std::vector<std::string> &variables) const
{
	std::map<std::string, std::string>::const_iterator it;
	int i;

	variables.resize(m_preprocessorVariables.size());

	for (i = 0, it = m_preprocessorVariables.begin() ; it != m_preprocessorVariables.end() ; it++, i++)
		variables[i] = it->first;
}

void ShellInstance::clearPreprocessorVariables()
{
	m_preprocessorVariables.clear();
}

//////////////////////////////////////

void ShellMainLoop(const std::string &shellPrefix, const std::string &prefix, bool exitOnError)
{
	std::string line;
	ShellInstance *pInst = ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	while (!pInst->isExitRequested() && pIOSys->readInputLine(prefix, line))
	{
		if (pIOSys->echo())
			pIOSys->print("%s%s", shellPrefix.c_str(), line.c_str());

		std::vector<std::string> arguments;

		IOSystem::splitLine(line, arguments);

		pInst->setInterrupted(false);
		if (!pInst->processCommandLine(arguments))
		{
			if (exitOnError)
				pInst->setErrorExit();

			pIOSys->flushExtraInputLines(); // make sure we don't process any extra lines
		}
		pInst->appendCommandLine(line);
	}
	//std::cout << "Exiting..." << std::endl;
}

#ifdef SHELLP_CONFIG_QT4

int qtMain(const std::string &title, const std::string &shellPrefix, int argc, char **argv)
{
	QApplication app(argc, argv);
	ShellMainWin mainWin(title, shellPrefix);

	ShellInstance *pInst = ShellInstance::getInstance();
	QtIOSystem *pQtIOSys = new QtIOSystem(&mainWin);

	pInst->setIOSystem(pQtIOSys);

	mainWin.startThread();

	pInst->setInteractive(true);
	pInst->onGUIThreadStartup();

	int status = app.exec();

	delete pQtIOSys;

	return status;
}

#endif // SHELLP_CONFIG_QT4

#ifdef SHELLP_CONFIG_NCURSES
void ShellCtrlCHandler(int val)
{
	ShellInstance *pInst = ShellInstance::getInstance();
	IOSystem *pIOSys = pInst->getIOSystem();

	pIOSys->writeOutputLine("Requesting interrupt of current calculation");
	pInst->setInterrupted();
}
#endif // SHELLP_CONFIG_NCURSES

int ShellInstance::run(const std::string &title, const std::string &shellPrefix, int argc, char **argv)
{
	if (argc > 2)
	{
		std::cerr << std::endl << "Usage: " << argv[0] << " [filename]" << std::endl << std::endl;
		std::cerr <<              "       Use '-' to specify standard input" << std::endl << std::endl;
		return -1;
	}

	std::string prefix;
	bool exitOnError = true;
	bool interactive = false;
	
	IOSystem *pIOSys = 0;

	if (argc == 2) // file name specified, try to open it
	{
		std::string arg(argv[1]);

		if (arg == std::string("-"))
		{
			StdIOSystem *pStdIOSys = new StdIOSystem(); 
			
			pIOSys = pStdIOSys;
		}
		else
		{
			FileIOSystem *pFileIOSys = new FileIOSystem();

			if (!pFileIOSys->init(argv[1]))
			{
				std::cerr << "ERROR: " << pFileIOSys->getErrorString() << std::endl;
				delete pFileIOSys;
				return -1;
			}
			pIOSys = pFileIOSys;
		}
	}
	else // No filename given, check if input is redirected
	{
#ifndef WIN32
#ifdef SHELLP_CONFIG_QT4
		if (getenv("DISPLAY") != 0)
			return qtMain(title, shellPrefix, argc, argv);
#endif // SHELLP_CONFIG_QT4

		if (ttyname(0) == 0) // input is not a tty, so no interactive input can be given
		{
			StdIOSystem *pStdIOSys = new StdIOSystem(); 
			
			pIOSys = pStdIOSys;
		}
		else // Input is a tty, interactivity possible
		{
			NCursesIOSystem *pNCursIOSys = new NCursesIOSystem();

			if (!pNCursIOSys->init(true))
			{
				std::cerr << "WARNING: Can't use ncurses interactive interface: " << pNCursIOSys->getErrorString() << std::endl;
#ifdef SHELLP_CONFIG_QT4
				std::cerr << "         Trying Qt interactive interface" << std::endl;
				delete pNCursIOSys;
				return qtMain(title, shellPrefix, argc, argv);
#else
				delete pNCursIOSys;
				return -1;
#endif // SHELLP_CONFIG_QT4
			}

#ifdef SHELLP_CONFIG_NCURSES
			// Install Ctrl-C handler
			signal(SIGINT, ShellCtrlCHandler);
#endif // SHELLP_CONFIG_NCURSES

			pIOSys = pNCursIOSys;

			prefix = shellPrefix;
			exitOnError = false;
			interactive = true;
		}
#else
#ifdef SHELLP_CONFIG_QT4
		return qtMain(title, shellPrefix, argc, argv);
#else
		std::cerr << "Qt4 was not available at compilation time, no interactive interface is available." << std::endl;
		return -1;
#endif // SHELLP_CONFIG_QT4
#endif // !WIN32
	}

	ShellInstance *pInst = ShellInstance::getInstance();
	
	pInst->setIOSystem(pIOSys);
	pInst->setInteractive(interactive);
	
	ShellMainLoop(shellPrefix, prefix, exitOnError);

	int returnStatus = 0;

	if (pInst->isErrorExit())
		returnStatus = -1;
	
	delete pIOSys;
	
	return returnStatus;
}

} // end namespace

